set B_WP27=2.7.10.3
set B_WP35=3.5.2.3
set B_WINPYTHON27=D:\wjp\WinPython-64bit-%B_WP27%
set B_WINPYTHON35=D:\wjp\WinPython-64bit-%B_WP35%
set B_MATLAB_ROOT=C:\Program Files\MATLAB\R2015a
set B_VC=C:\Program Files (x86)\Microsoft Visual Studio 14.0\VC\bin\amd64
set B_BV=1_62
set B_BOOST=D:\wjp\boost_%B_BV%_0

set B_VCREDIST=D:\wjp\vs2015u3_redist\vc_redist.x64.exe

set B_README_WP27=C:\WinPython-64bit-%B_WP27%\python-%B_WP27:~0,-2%.amd64\Lib\site-packages
set B_README_WP35=C:\WinPython-64bit-%B_WP35%\python-%B_WP35:~0,-2%.amd64\Lib\site-packages

set B_RELEASE_TAG=v1.8.3
set B_RELEASE=1.8.3
